@@ -38,16 +38,18 @@
 #include <sys/stat.h>
 #include <unistd.h>
 
+#include <gconf/gconf-client.h>
 #include <libgnomevfs/gnome-vfs.h>
 #include <libgnomevfs/gnome-vfs-mime.h>
 
-#include <libgnomevfs/gnome-vfs-i18n.h>
 #include <libgnomevfs/gnome-vfs-method.h>
 #include <libgnomevfs/gnome-vfs-module.h>
 #include <libgnomevfs/gnome-vfs-module-shared.h>
 #include <libgnomevfs/gnome-vfs-module-callback-module-api.h>
 #include <libgnomevfs/gnome-vfs-standard-callbacks.h>
 
+#include <glib/gi18n-lib.h>
+
 #include <libsmbclient.h>
 
 int smbc_remove_unused_server(SMBCCTX * context, SMBCSRV * srv);
@@ -94,9 +96,14 @@ static GHashTable *user_cache = NULL;
 #define CACHE_REAP_TIMEOUT (30 * 60)
 static guint cache_reap_timeout = 0;
 
+/* We load a default workgroup from gconf */
+#define PATH_GCONF_GNOME_VFS_SMB_WORKGROUP "/system/smb/workgroup"
+
 /* The magic "default workgroup" hostname */
 #define DEFAULT_WORKGROUP_NAME "X-GNOME-DEFAULT-WORKGROUP"
 
+/* Guest logins use: */
+#define GUEST_LOGIN "guest"
 
 /* 5 minutes before we re-read the workgroup cache again */
 #define WORKGROUP_CACHE_TIMEOUT (5*60)
@@ -107,9 +114,9 @@ static time_t workgroups_timestamp = 0;
 /* Authentication ----------------------------------------------------------- */
 
 #define SMB_AUTH_STATE_PREFILLED	0x00000010 	/* Have asked gnome-auth for prefilled auth */
-#define SMB_AUTH_STATE_PROMPTED		0x00000020 	/* Have asked gnome-auth for to prompt user */
+#define SMB_AUTH_STATE_GUEST		0x00000020 	/* Have tried 'guest' authentication */
+#define SMB_AUTH_STATE_PROMPTED		0x00000040 	/* Have asked gnome-auth for to prompt user */
 
-/* TODO: Move this to the top of the file */
 typedef struct _SmbAuthContext {
 	
 	GnomeVFSURI *uri;		/* Uri being worked with. Does not own this URI */
@@ -125,11 +132,17 @@ typedef struct _SmbAuthContext {
 	
 	/* Used in chat between perform_authentication and auth_callback */
 	gboolean auth_called;		/* Set by auth_callback */
+	gboolean preset_user;		/* Set when we have a preset user from the URI */
 	gchar *for_server;		/* Set by auth_callback */
 	gchar *for_share;		/* Set by auth_callback */
 	gchar *use_user;		/* Set by perform_authentication */
 	gchar *use_domain;		/* Set by perform_authentication */
 	gchar *use_password;		/* Set by perform_authentication */
+
+	gboolean cache_added;           /* Set when cache is added to during authentication */
+	gboolean cache_used;            /* Set when cache is used during authentication */
+
+	guint prompt_flags;           	/* Various URI flags set by initial_authentication */
 	
 } SmbAuthContext;
 
@@ -138,9 +151,6 @@ static int  perform_authentication (SmbAuthContext *actx);
 
 static SmbAuthContext *current_auth_context = NULL;
 
-/* Used to detect failed logins */
-static gboolean cache_access_failed = FALSE;
-
 static void auth_callback (const char *server_name, const char *share_name,
 		     	   char *domain, int domainmaxlen,
 		     	   char *username, int unmaxlen,
@@ -345,8 +355,11 @@ add_cached_server (SMBCCTX *context, SMBCSRV *new,
 {
 	SmbServerCacheEntry *entry = NULL;
 
-	DEBUG_SMB(("add_cached_server: server: %s, share: %s, domain: %s, user: %s\n",
-		   server_name, share_name, domain, username));
+	DEBUG_SMB(("[auth] adding cached server: server: %s, share: %s, domain: %s, user: %s\n",
+		   server_name ? server_name : "", 
+		   share_name ? share_name : "",
+		   domain ? domain : "",
+		   username ? username : ""));
 	
 	schedule_cache_reap ();
 	
@@ -361,19 +374,23 @@ add_cached_server (SMBCCTX *context, SMBCSRV *new,
 	entry->last_time = time (NULL);
 
 	g_hash_table_insert (server_cache, entry, entry);
-
-	cache_access_failed = FALSE;
+	current_auth_context->cache_added = TRUE;
 	return 0;
 }
 
 static SMBCSRV *
-get_cached_server (SMBCCTX * context,
-		   const char *server_name, const char *share_name,
-		   const char *domain, const char *username)
+find_cached_server (const char *server_name, const char *share_name,
+                    const char *domain, const char *username)
 {
 	SmbServerCacheEntry entry;
 	SmbServerCacheEntry *res;
 
+	DEBUG_SMB(("find_cached_server: server: %s, share: %s, domain: %s, user: %s\n", 
+		   server_name ? server_name : "", 
+		   share_name ? share_name : "",
+		   domain ? domain : "", 
+		   username ? username : ""));
+
 	entry.server_name = (char *)server_name;
 	entry.share_name = (char *)share_name;
 	entry.domain = (char *)domain;
@@ -382,11 +399,30 @@ get_cached_server (SMBCCTX * context,
 	res = g_hash_table_lookup (server_cache, &entry);
 
 	if (res != NULL) {
-		cache_access_failed = FALSE;
 		res->last_time = time (NULL);
 		return res->server;
+	} 
+
+	return NULL;
+}
+
+static SMBCSRV *
+get_cached_server (SMBCCTX * context,
+		   const char *server_name, const char *share_name,
+		   const char *domain, const char *username)
+{
+	SMBCSRV *srv;
+	
+	srv = find_cached_server (server_name, share_name, domain, username);
+	if (srv != NULL) {
+		DEBUG_SMB(("got cached server: server: %s, share: %s, domain: %s, user: %s\n",
+		   	   server_name ? server_name : "", 
+			   share_name ? share_name : "",
+			   domain ? domain : "", 
+			   username ? username : ""));
+		current_auth_context->cache_used = TRUE;
+		return srv;
 	}
-	cache_access_failed = TRUE;
 	return NULL;
 }
 
@@ -585,6 +621,8 @@ static gboolean
 try_init (void)
 {
 	char *path;
+	GConfClient *gclient;
+	gchar *workgroup;
 	struct stat statbuf;
 
 	LOCK_SMB();
@@ -613,6 +651,20 @@ try_init (void)
 		smb_context->callbacks.remove_cached_srv_fn = remove_cached_server;
 		smb_context->callbacks.purge_cached_fn      = purge_cached;
 
+		gclient = gconf_client_get_default ();
+		if (gclient) {
+			workgroup = gconf_client_get_string (gclient, 
+					PATH_GCONF_GNOME_VFS_SMB_WORKGROUP, NULL);
+
+			/* libsmbclient frees this on it's own, so make sure 
+			 * to use simple system malloc */
+			if (workgroup && workgroup[0])
+				smb_context->workgroup = strdup (workgroup);
+			
+			g_free (workgroup);
+			g_object_unref (gclient);
+		}
+
 		if (!smbc_init_context (smb_context)) {
 			smbc_free_context (smb_context, FALSE);
 			smb_context = NULL;
@@ -648,17 +700,26 @@ try_init (void)
 
 /* Authentication ----------------------------------------------------------- */
 
+static const gchar*
+get_auth_display_share (SmbAuthContext *actx)
+{
+        return string_compare (actx->for_share, "IPC$") ? 
+                        NULL : actx->for_share;
+}
+
 static char*
 get_auth_display_uri (SmbAuthContext *actx, gboolean machine)
 {
         if (actx->uri != NULL && !machine)
                 return gnome_vfs_uri_to_string (actx->uri, 0);
-         else
+        else {
+                const gchar *share = get_auth_display_share (actx);
                 return g_strdup_printf ("smb://%s%s%s%s", 
                                         actx->for_server ? actx->for_server : "", 
                                         actx->for_server ? "/" : "",
-                                        actx->for_share && !machine ? actx->for_share : "",
-                                        actx->for_share && !machine ? "/" : "");
+                                        share && !machine ? share : "",
+                                        share && !machine ? "/" : "");
+        }
 }
 
 static void
@@ -672,7 +733,10 @@ update_user_cache (SmbAuthContext *actx, gboolean with_share)
         key = g_strdup_printf ("%s/%s", actx->for_server, with_share ? actx->for_share : "");
         user = (SmbCachedUser*)g_hash_table_lookup (user_cache, key);
         
-        DEBUG_SMB(("[auth] Saved in cache: %s = %s:%s@%s\n", key, actx->use_user, actx->use_domain, actx->use_password));
+        DEBUG_SMB(("[auth] Saved in cache: %s = %s:%s@%s\n", key,
+		   actx->use_user ? actx->use_user : "", 
+		   actx->use_domain ? actx->use_domain : "", 
+		   actx->use_password ? actx->use_password : ""));
         
         if (!user) {
                 user = g_new0 (SmbCachedUser, 1);
@@ -689,7 +753,7 @@ update_user_cache (SmbAuthContext *actx, gboolean with_share)
 }
 
 static gboolean
-lookup_user_cache (SmbAuthContext *actx, gboolean with_share, gboolean with_user)
+lookup_user_cache (SmbAuthContext *actx, gboolean with_share)
 {
         SmbCachedUser *user;
         gchar *key;
@@ -700,16 +764,22 @@ lookup_user_cache (SmbAuthContext *actx, gboolean with_share, gboolean with_user
         user = (SmbCachedUser*)g_hash_table_lookup (user_cache, key);
         g_free (key);
        
-        /* If the caller specified we already have a user name double check that... */
-        if (with_user && (!string_compare (user->username, actx->use_user) ||
-                          !string_compare (user->domain, actx->use_domain)))
-                return FALSE;
-                        
         if (user) {
+                /* If we already have a user name or domain double check that... */
+		if (!(actx->prompt_flags & GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_USERNAME) &&
+		    !string_compare(user->username, actx->use_user))
+			return FALSE;
+		if (!(actx->prompt_flags & GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_DOMAIN) &&
+		    !string_compare(user->domain, actx->use_domain))
+			return FALSE;
+
                 actx->use_user = string_realloc (actx->use_user, user->username);
                 actx->use_domain = string_realloc (actx->use_domain, user->domain);
                 actx->use_password = string_realloc (actx->use_password, user->password);
-                DEBUG_SMB(("[auth] Looked up in cache: %s:%s@%s\n", actx->use_user, actx->use_domain, actx->use_password));
+                DEBUG_SMB(("[auth] Looked up in cache: %s:%s@%s\n", 
+			   actx->use_user ? actx->use_user : "",
+			   actx->use_domain ? actx->use_domain : "",
+			   actx->use_password ? actx->use_password : ""));
                 return TRUE;
         }
         
@@ -725,12 +795,13 @@ initial_authentication (SmbAuthContext *actx)
 	SmbServerCacheEntry server_lookup;
 	SmbServerCacheEntry *server;
         gboolean found_user = FALSE;
-        gboolean preset_user = FALSE;
 	char *tmp;
 
 	DEBUG_SMB(("[auth] Initial authentication lookups\n"));
 
 	toplevel_uri =	(GnomeVFSToplevelURI*)actx->uri;
+	actx->prompt_flags = GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_USERNAME |
+		    	     GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_DOMAIN;
 	
 	/* Try parsing a user and domain out of the URI */
 	if (toplevel_uri && toplevel_uri->user_name != NULL && 
@@ -743,34 +814,54 @@ initial_authentication (SmbAuthContext *actx)
 					    	  	      tmp - toplevel_uri->user_name);
 			g_free (actx->use_user);
 			actx->use_user = string_dup_nzero (tmp + 1);
-			DEBUG_SMB(("[auth] User from URI: %s@%s\n", actx->use_user, actx->use_domain));
+			DEBUG_SMB(("[auth] User from URI: %s@%s\n", 
+				   actx->use_user ? actx->use_user : "", 
+				   actx->use_domain ? actx->use_domain : ""));
 		} else {
 			g_free (actx->use_user);
 			actx->use_user = string_dup_nzero (toplevel_uri->user_name);
 			g_free (actx->use_domain);
 			actx->use_domain = NULL;
-			DEBUG_SMB(("[auth] User from URI: %s\n", actx->use_user));
+			DEBUG_SMB(("[auth] User from URI: %s\n", 
+				   actx->use_user ? actx->use_user : ""));
+		}
+		
+		if (actx->use_user != NULL) {
+			actx->preset_user = TRUE;
+			actx->prompt_flags &= ~GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_USERNAME;
+		}
+		if (actx->use_domain != NULL) {
+			actx->prompt_flags &= ~GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_DOMAIN;
 		}
-                preset_user = TRUE;
 	} 
 
-        if (lookup_user_cache (actx, TRUE, preset_user) ||
-            lookup_user_cache (actx, FALSE, preset_user))
+        if (lookup_user_cache (actx, TRUE) ||
+            lookup_user_cache (actx, FALSE))
                 found_user = TRUE;
 
-        if (found_user || preset_user) {
-        	/* Lookup the password in our internal cache */
+        if (found_user || actx->preset_user) {
+        	/* Lookup the server in our internal cache */
         	server_lookup.server_name = (char*)actx->for_server;
         	server_lookup.share_name = (char*)actx->for_share;
         	server_lookup.username = (char*)actx->use_user;
         	server_lookup.domain = (char*)actx->use_domain;
         		
         	server = g_hash_table_lookup (server_cache, &server_lookup);
-        	if (server != NULL) {
-        		/* Server is in cache already, no need to get password */
-        		g_free (actx->use_password);
-        		actx->use_password = g_strdup ("");
-        		DEBUG_SMB(("[auth] Using login info for '%s@%s' from our internal cache\n", actx->use_user, actx->use_domain));
+        	if (server == NULL) {
+                 
+                        /* If a blank user, try looking up 'guest' */
+                        if (!actx->use_user) {
+                                server_lookup.username = GUEST_LOGIN;
+                                server_lookup.domain = NULL;
+                                server = g_hash_table_lookup (server_cache, &server_lookup);
+                        }
+                }
+                
+                /* Server is in cache already */
+                if (server != NULL) {
+        		DEBUG_SMB(("[auth] Using connection for '%s@%s' from our server cache\n", 
+				   actx->use_user ? actx->use_user : "", 
+				   actx->use_domain ? actx->use_domain : ""));
                         found_user = TRUE;
         	}
         }
@@ -786,6 +877,7 @@ prefill_authentication (SmbAuthContext *actx)
 	GnomeVFSModuleCallbackFillAuthenticationIn in_args;
 	GnomeVFSModuleCallbackFillAuthenticationOut out_args;
 	gboolean invoked;
+	gboolean filled = FALSE;
 
 	g_return_val_if_fail (actx != NULL, FALSE);
 	g_return_val_if_fail (actx->for_server != NULL, FALSE);	
@@ -794,14 +886,15 @@ prefill_authentication (SmbAuthContext *actx)
 	in_args.uri = get_auth_display_uri (actx, FALSE);
 	in_args.protocol = "smb";
 	in_args.server = (char*)actx->for_server;
-	in_args.object = (char*)actx->for_share;
+	in_args.object = (char*)get_auth_display_share (actx);
 	in_args.username = (char*)actx->use_user;
 	in_args.domain = (char*)actx->use_domain;
 	in_args.port = actx->uri ? ((GnomeVFSToplevelURI*)actx->uri)->host_port : 0;
 
 	memset (&out_args, 0, sizeof (out_args));
 
-	DEBUG_SMB(("[auth] Trying to prefill credentials for: %s\n", in_args.uri));
+	DEBUG_SMB(("[auth] Trying to prefill credentials for: %s\n", 
+		   in_args.uri ? in_args.uri : ""));
 
 	invoked = gnome_vfs_module_callback_invoke
 		              (GNOME_VFS_MODULE_CALLBACK_FILL_AUTHENTICATION,
@@ -828,39 +921,50 @@ prefill_authentication (SmbAuthContext *actx)
 
                 memset (&out_args, 0, sizeof (out_args));
 
-                DEBUG_SMB(("[auth] Trying to prefill server credentials for: %s\n", in_args.uri));
+                DEBUG_SMB(("[auth] Trying to prefill server credentials for: %s\n", 
+			   in_args.uri ? in_args.uri : ""));
 
                 invoked = gnome_vfs_module_callback_invoke
                                         (GNOME_VFS_MODULE_CALLBACK_FILL_AUTHENTICATION,
                                         &in_args, sizeof (in_args),
-                                        &out_args, sizeof (out_args));                               
+                                        &out_args, sizeof (out_args));                              
         }
         
 	if (invoked && out_args.valid) {
-		g_free (actx->use_user);
-		actx->use_user = string_dup_nzero (out_args.username);
-		g_free (actx->use_domain);
-		actx->use_domain = string_dup_nzero (out_args.domain);
-		g_free (actx->use_password);
-		actx->use_password = g_strdup (out_args.password);
-		DEBUG_SMB(("[auth] Prefilled credentials: %s@%s:%s\n", actx->use_user, actx->use_domain, actx->use_password));
+		/* When a preset user make sure we stick to this login */
+		if (!actx->preset_user || string_compare (actx->use_user, out_args.username)) {
+			
+			g_free (actx->use_user);
+			actx->use_user = string_dup_nzero (out_args.username);
+			g_free (actx->use_domain);
+			actx->use_domain = string_dup_nzero (out_args.domain);
+			g_free (actx->use_password);
+			actx->use_password = g_strdup (out_args.password);
+			
+			filled = TRUE;
+			DEBUG_SMB(("[auth] Prefilled credentials: %s@%s:%s\n", 
+				   actx->use_user ? actx->use_user : "", 
+				   actx->use_domain ? actx->use_domain : "", 
+				   actx->use_password ? actx->use_password : ""));
+		}
 	} 
 	
 	g_free (out_args.username);
 	g_free (out_args.domain);
 	g_free (out_args.password);
 
-	return invoked && out_args.valid;
+	return filled;
 }
 
 static gboolean
-prompt_authentication (SmbAuthContext *actx)
+prompt_authentication (SmbAuthContext *actx,
+		       gboolean       *cancelled)
 {
 	/* IMPORTANT: We are NOT in the lock at this point */
 
 	GnomeVFSModuleCallbackFullAuthenticationIn in_args;
 	GnomeVFSModuleCallbackFullAuthenticationOut out_args;
-	gboolean invoked, cancelled = FALSE;
+	gboolean invoked;
 	
 	g_return_val_if_fail (actx != NULL, FALSE);
 	g_return_val_if_fail (actx->for_server != NULL, FALSE);
@@ -870,24 +974,28 @@ prompt_authentication (SmbAuthContext *actx)
 	in_args.flags = GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_PASSWORD | GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_SAVING_SUPPORTED;
 	if (actx->state & SMB_AUTH_STATE_PROMPTED)
 		in_args.flags |= GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_PREVIOUS_ATTEMPT_FAILED;
-	if (actx->uri && ((GnomeVFSToplevelURI*)actx->uri)->user_name == NULL) {
-		in_args.flags |= GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_USERNAME |
-	  			 GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_DOMAIN;
-	}
+	in_args.flags |= actx->prompt_flags;
 
 	in_args.uri = get_auth_display_uri (actx, FALSE);
 	in_args.protocol = "smb";
 	in_args.server = (char*)actx->for_server;
-	in_args.object = (char*)actx->for_share;
+	in_args.object = (char*)get_auth_display_share (actx);
 	in_args.username = (char*)actx->use_user;
 	in_args.domain = (char*)actx->use_domain;
 	in_args.port = actx->uri ? ((GnomeVFSToplevelURI*)actx->uri)->host_port : 0;
 
-	/* TODO: set default_user & default_domain? */
+	in_args.default_user = actx->use_user;
+	if (string_compare (in_args.default_user, GUEST_LOGIN))
+		in_args.default_user = NULL;
+	if (!in_args.default_user)
+		in_args.default_user = (char*)g_get_user_name ();
+	
+	in_args.default_domain = actx->use_domain ? actx->use_domain : smb_context->workgroup;
 	
 	memset (&out_args, 0, sizeof (out_args));
 
-	DEBUG_SMB(("[auth] Prompting credentials for: %s\n", in_args.uri));
+	DEBUG_SMB(("[auth] Prompting credentials for: %s\n", 
+		   in_args.uri ? in_args.uri : ""));
 
 	invoked = gnome_vfs_module_callback_invoke
 		(GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION,
@@ -895,7 +1003,6 @@ prompt_authentication (SmbAuthContext *actx)
 		 &out_args, sizeof (out_args));
 
 	if (invoked) {
-		cancelled = out_args.abort_auth;
                 if (in_args.flags & GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_USERNAME) {
                      g_free (actx->use_user);
                        actx->use_user = string_dup_nzero (out_args.username);
@@ -909,9 +1016,14 @@ prompt_authentication (SmbAuthContext *actx)
 		g_free (actx->keyring);
 		actx->save_auth = out_args.save_password;
 		actx->keyring = actx->save_auth && out_args.keyring ? g_strdup (out_args.keyring) : NULL;
-		DEBUG_SMB(("[auth] Prompted credentials: %s@%s:%s\n", actx->use_user, actx->use_domain, actx->use_password));
+		DEBUG_SMB(("[auth] Prompted credentials: %s@%s:%s\n", 
+			    actx->use_user ? actx->use_user : "", 
+			    actx->use_domain ? actx->use_domain : "",
+			    actx->use_password ? actx->use_password : ""));
 	} 
-	
+
+	*cancelled = out_args.abort_auth;
+
 	actx->state |= SMB_AUTH_STATE_PROMPTED;
 
 	g_free (out_args.username);
@@ -921,7 +1033,7 @@ prompt_authentication (SmbAuthContext *actx)
 
 	g_free (in_args.uri);
 
-	return invoked && !cancelled;
+	return invoked && !*cancelled;
 }
 
 static void
@@ -954,8 +1066,13 @@ save_authentication (SmbAuthContext *actx)
 	in_args.password = (char*)actx->use_password;
 
 	memset (&out_args, 0, sizeof (out_args));
-	DEBUG_SMB(("[auth] Saving credentials: %s = %s@%s:%s\n", in_args.uri, actx->use_user, actx->use_domain, actx->use_password));
-	DEBUG_SMB(("[auth] Keyring: %s\n", actx->keyring));
+	DEBUG_SMB(("[auth] Saving credentials: %s = %s@%s:%s\n", 
+		   in_args.uri ? in_args.uri : "", 
+		   actx->use_user ? actx->use_user : "", 
+		   actx->use_domain ? actx->use_domain : "",
+		   actx->use_password ? actx->use_password : ""));
+	DEBUG_SMB(("[auth] Keyring: %s\n", 
+		   actx->keyring ? actx->keyring : ""));
 
 	invoked = gnome_vfs_module_callback_invoke
                                 (GNOME_VFS_MODULE_CALLBACK_SAVE_AUTHENTICATION,
@@ -970,7 +1087,7 @@ save_authentication (SmbAuthContext *actx)
         in_args.keyring = (char*)actx->keyring;
         in_args.protocol = "smb";
         in_args.server = (char*)actx->for_server;
-        in_args.object = (char*)actx->for_share;
+        in_args.object = (char*)get_auth_display_share (actx);
         in_args.port = actx->uri ? ((GnomeVFSToplevelURI*)actx->uri)->host_port : 0;
         in_args.authtype = NULL;
         in_args.username = (char*)actx->use_user;
@@ -978,8 +1095,13 @@ save_authentication (SmbAuthContext *actx)
         in_args.password = (char*)actx->use_password;
         
         memset (&out_args, 0, sizeof (out_args));
-        DEBUG_SMB(("[auth] Saving credentials: %s = %s@%s:%s\n", in_args.uri, actx->use_user, actx->use_domain, actx->use_password));
-        DEBUG_SMB(("[auth] Keyring: %s\n", actx->keyring));
+        DEBUG_SMB(("[auth] Saving credentials: %s = %s@%s:%s\n", 
+		   in_args.uri ? in_args.uri : "", 
+		   actx->use_user ? actx->use_user : "", 
+		   actx->use_domain ? actx->use_domain : "", 
+		   actx->use_password ? actx->use_password : ""));
+        DEBUG_SMB(("[auth] Keyring: %s\n", 
+		   actx->keyring ? actx->keyring : ""));
 
         invoked = gnome_vfs_module_callback_invoke
                                 (GNOME_VFS_MODULE_CALLBACK_SAVE_AUTHENTICATION,
@@ -1054,13 +1176,12 @@ init_authentication (SmbAuthContext *actx, GnomeVFSURI *uri)
 	DEBUG_SMB(("[auth] Initializing Authentication\n"));
 	memset (actx, 0, sizeof(*actx));
 	actx->uri = uri;
-	cache_access_failed = FALSE;
 }
 
 static int
 perform_authentication (SmbAuthContext *actx)
 {
-	gboolean cont, auth_failed = FALSE;
+	gboolean cont, auth_failed = FALSE, auth_cancelled = FALSE;
 	int ret = -1;
 	
 	/* IMPORTANT: We are IN the lock at this point */
@@ -1085,8 +1206,6 @@ perform_authentication (SmbAuthContext *actx)
 		return -1;
 	}
 	
-	/* TODO: What about cache_access_failed? */
-	
 	actx->passes++;
 
 	/* First pass */
@@ -1113,7 +1232,15 @@ perform_authentication (SmbAuthContext *actx)
 			DEBUG_SMB(("[auth] Operation successful\n"));
 			save_authentication (actx);
 			ret = 0;
-	
+
+		/* If authentication failed, but we already have a connection 
+		   ... and access failed on a file, then we return the error */
+		} else if ((actx->cache_used && !actx->cache_added) && 
+			(!actx->uri || smb_uri_type (actx->uri) == SMB_URI_SHARE_FILE)) {
+
+			DEBUG_SMB(("[auth] Not reauthenticating a open connection.\n"));
+			ret = -1;
+
 		/* A failed authentication */
 		} else if (actx->auth_called) {
 			
@@ -1125,14 +1252,28 @@ perform_authentication (SmbAuthContext *actx)
 			cont = FALSE;
 			
 			UNLOCK_SMB();
-
+			
+				/* Do we have gnome-keyring credentials for this? */
 				if (!(actx->state & SMB_AUTH_STATE_PREFILLED)) {
 					actx->state |= SMB_AUTH_STATE_PREFILLED;
 					cont = prefill_authentication (actx);
 				}
-				
+
+				/* Then we try a guest credentials... */
+				if (!cont && !actx->preset_user && !(actx->state & SMB_AUTH_STATE_GUEST)) {
+					g_free (actx->use_user);
+					actx->use_user = strdup(GUEST_LOGIN);
+					g_free (actx->use_domain);
+					actx->use_domain = NULL;
+					g_free (actx->use_password);
+					actx->use_password = strdup("");
+					actx->state |= SMB_AUTH_STATE_GUEST;
+					cont = TRUE;
+				}
+
+				/* And as a last step, prompt */
 				if (!cont)
-					cont = prompt_authentication (actx);
+					cont = prompt_authentication (actx, &auth_cancelled);
 				
 			LOCK_SMB();
 			
@@ -1143,9 +1284,16 @@ perform_authentication (SmbAuthContext *actx)
 			if (cont)
 				ret = 1;
 			else {
-				DEBUG_SMB(("[auth] Authentication cancelled by user.\n"));
-				actx->res = GNOME_VFS_ERROR_CANCELLED;
 				ret = -1;
+
+				if (auth_cancelled) {
+					DEBUG_SMB(("[auth] Authentication cancelled by user.\n"));
+					actx->res = GNOME_VFS_ERROR_CANCELLED;
+				} else {
+					DEBUG_SMB(("[auth] Authentication failed with result %s.\n",
+						  gnome_vfs_result_to_string (actx->res)));
+					/* Note that we leave actx->res set to whatever it was set to */
+				}
 			}
 					
 		/* Weird, don't want authentication, but failed */
@@ -1169,9 +1317,11 @@ auth_callback (const char *server_name, const char *share_name,
 {
 	/* IMPORTANT: We are IN the global lock */
 	SmbAuthContext *actx;
+	SMBCSRV *server;
 	
 	DEBUG_SMB (("[auth] auth_callback called: server: %s share: %s\n",
-		    server_name, share_name));
+		    server_name ? server_name : "", 
+		    share_name ? share_name : ""));
 
 	g_return_if_fail (current_auth_context != NULL);
 	actx = current_auth_context;
@@ -1186,27 +1336,51 @@ auth_callback (const char *server_name, const char *share_name,
 	g_free (actx->for_server);
 	actx->for_server = string_dup_nzero (server_name);
 	g_free (actx->for_share);
-	actx->for_share = !share_name || strcmp (share_name,"IPC$") == 0 
-				? NULL : string_dup_nzero (share_name);
+	actx->for_share = string_dup_nzero (share_name);
 
 	/* The first pass, try the cache, fill in anything we know */
 	if (actx->passes == 1)
 		initial_authentication (actx);
 	
-	/* If we have a valid user and password then go for it */
-	if (actx->use_user && actx->use_password) {
-		DEBUG_SMB(("[auth] Using credentials: %s@%s:%s\n", actx->use_user, actx->use_domain, actx->use_password));
+	/* If we have a valid user then go for it */
+	if (actx->use_user) {
 		strncpy (username_out, actx->use_user, unmaxlen);
-		strncpy (password_out, actx->use_password, pwmaxlen);
+                strncpy (password_out, actx->use_password ? actx->use_password : "", pwmaxlen);
 		if (actx->use_domain)
 			strncpy (domain_out, actx->use_domain, domainmaxlen);
-
-	/* We have no credentials ... */			
+		DEBUG_SMB(("[auth] Using credentials: %s:%s@%s\n", 
+			   username_out ? username_out : "", 
+			   password_out ? password_out : "", 
+			   domain_out ? domain_out : ""));
+	
+	/* We have no credentials ... */
 	} else {
-		DEBUG_SMB(("[auth] No credentials, returning null values\n"));
+		g_assert (!actx->preset_user);
+		
+		if (actx->passes == 1)
+			DEBUG_SMB(("[auth] No credentials, trying anonymous user login\n"));
+		else
+			DEBUG_SMB(("[auth] No credentials, returning null values\n"));
+		
 		strncpy (username_out, "", unmaxlen);
 		strncpy (password_out, "", pwmaxlen);
-		strncpy (domain_out, "", domainmaxlen);
+	}
+
+	/* Put in the default workgroup if none specified */
+	if (domain_out[0] == 0 && smb_context->workgroup)
+		strncpy (domain_out, smb_context->workgroup, domainmaxlen);
+
+	/* 
+	 * If authentication is requested a second time on a server we've 
+	 * already cached, then obviously it was invalid. Remove it. Yes, 
+	 * this doesn't make much sense, but such is life with libsmbclient.
+	 */
+	if ((actx->state & SMB_AUTH_STATE_PROMPTED) && actx->res != GNOME_VFS_OK) {
+		server = find_cached_server (server_name, share_name, domain_out, username_out);
+		if (server) {
+			DEBUG_SMB (("[auth] auth_callback. Remove the wrong server entry from server_cache.\n"));
+			g_hash_table_foreach_remove (server_cache, remove_server, server);
+		}
 	}
 }
 
@@ -1393,7 +1567,11 @@ do_close (GnomeVFSMethod *method,
 
 		/* Important: perform_authentication leaves and re-enters the lock! */
 		while (perform_authentication (&actx) > 0) {
+#ifdef HAVE_SAMBA_OLD_CLOSE
 			r = smb_context->close (smb_context, handle->file);
+#else
+			r = smb_context->close_fn (smb_context, handle->file);
+#endif
 			actx.res = (r >= 0) ? GNOME_VFS_OK : gnome_vfs_result_from_errno ();
 		}
 
@@ -1414,7 +1592,7 @@ do_read (GnomeVFSMethod *method,
 	 GnomeVFSContext *context)
 {
 	FileHandle *handle = (FileHandle *)method_handle;
-	GnomeVFSResult res = GNOME_VFS_OK;;
+	GnomeVFSResult res = GNOME_VFS_OK;
 	SmbAuthContext actx;
 	ssize_t n = 0;
 
@@ -2087,7 +2265,7 @@ do_seek (GnomeVFSMethod *method,
 static GnomeVFSResult
 do_tell (GnomeVFSMethod *method,
 		GnomeVFSMethodHandle *method_handle,
-		GnomeVFSFileOffset *offset_return)
+		GnomeVFSFileSize *offset_return)
 {
 	FileHandle *handle = (FileHandle *)method_handle;
 	SmbAuthContext actx;
